-------------------------------------------------------------------------------
-- audiotop
--
-- Author(s):     James Brennan and Jorgen Peddersen
-- Created:       Dec 2000
-- Last Modified: Dec 2000
-- 
-- Map audio device to the RAM controller.  See other files for descriptions.
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;

entity audiotop is
	port (
		clk: in STD_LOGIC;
		rstn: in STD_LOGIC;
		direction: in STD_LOGIC;
		speed: in STD_LOGIC_VECTOR(1 downto 0);	
		precision: in STD_LOGIC;
		volume: in STD_LOGIC_VECTOR(1 downto 0);
			
		CELeftn: out STD_LOGIC;
        CERightn: out STD_LOGIC;
        OELeftn: out STD_LOGIC;
        OERightn: out STD_LOGIC;
        WELeftn: out STD_LOGIC;
        WERightn: out STD_LOGIC;
        SRAMLeftAddr: out STD_LOGIC_VECTOR (18 downto 0);
        SRAMRightAddr: out STD_LOGIC_VECTOR (18 downto 0);
        SRAMLeftData: inout STD_LOGIC_VECTOR (15 downto 0);
        SRAMRightData: inout STD_LOGIC_VECTOR (15 downto 0);

		sdout: in STD_LOGIC;
		
		mclk: out STD_LOGIC;
        lrck: buffer STD_LOGIC;
        sclk: buffer STD_LOGIC;
        sdin: out STD_LOGIC;		

		recordStart: in STD_LOGIC;
		playStart: in STD_LOGIC;
		encode : in STD_LOGIC;		
		full: out STD_LOGIC;
      done: out STD_LOGIC;
		encodedone : out STD_LOGIC;
		doingstuff : out STD_LOGIC
);
end audiotop;

architecture audiotop_structure of audiotop is



component audio
    port (
        clk: in STD_LOGIC;
        rstn: in STD_LOGIC;
        sdout: in STD_LOGIC;
        recordStart: in STD_LOGIC;
        recordStop: in STD_LOGIC;
        playStart: in STD_LOGIC;
        playStop: in STD_LOGIC;
		direction: in STD_LOGIC;
		speed: in STD_LOGIC_VECTOR(1 downto 0);
		precision: in STD_LOGIC;
		volume: in STD_LOGIC_VECTOR(1 downto 0);
        dataIn: in STD_LOGIC_VECTOR(31 downto 0);
        mclk: out STD_LOGIC;
        lrck: buffer STD_LOGIC;
        sclk: buffer STD_LOGIC;
        sdin: out STD_LOGIC;
        recordAddress: out STD_LOGIC_VECTOR(18 downto 0);
        playAddress: out STD_LOGIC_VECTOR(18 downto 0);
        dataOut: out STD_LOGIC_VECTOR(31 downto 0);
        read: out STD_LOGIC;
        write: out STD_LOGIC;
        full: out STD_LOGIC;
        done: out STD_LOGIC
    );
end component;        

component sraminterface
    port (
        CLK: in STD_LOGIC;
        Resetn: in STD_LOGIC;
        doRead: in STD_LOGIC;
        doWrite: in STD_LOGIC;
        readAddr: in STD_LOGIC_VECTOR (18 downto 0);
        writeAddr: in STD_LOGIC_VECTOR (18 downto 0);
        readData: out STD_LOGIC_VECTOR (31 downto 0);
        writeData: in STD_LOGIC_VECTOR (31 downto 0);
        canRead: out STD_LOGIC;
        canWrite: out STD_LOGIC;
        CELeftn: out STD_LOGIC;
        CERightn: out STD_LOGIC;
        OELeftn: out STD_LOGIC;
        OERightn: out STD_LOGIC;
        WELeftn: out STD_LOGIC;
        WERightn: out STD_LOGIC;
        SRAMLeftAddr: out STD_LOGIC_VECTOR (18 downto 0);
        SRAMRightAddr: out STD_LOGIC_VECTOR (18 downto 0);
        SRAMLeftData: inout STD_LOGIC_VECTOR (15 downto 0);
        SRAMRightData: inout STD_LOGIC_VECTOR (15 downto 0)
    );
end component;

component encoder 
	port (
		     read : out std_logic;
           write : out std_logic;
			  inaddr : out std_logic_vector(18 downto 0);
	        outaddr : out std_logic_vector(18 downto 0);
           indata : in std_logic_vector(31 downto 0);  	-- ADPCM encrypted
           outdata : out std_logic_vector(31 downto 0);	-- wave sound file
			  canread : in std_logic;
			  canwrite : in std_logic;
           encode : in std_logic;
			  clk : in std_logic;
			  reset : in std_logic;
			  encodedone : out std_logic;
			  doingstuff : out std_logic
	);
end component;

component encodemux
	port (  encode : in std_logic;
           playStart : in std_logic;
			  recordStart : in std_logic;
			  reset : in std_logic;
           playAddress : in std_logic_vector(18 downto 0);
           recordAddress : in std_logic_vector(18 downto 0);
           playData : out std_logic_vector(31 downto 0);
           recordData : in std_logic_vector(31 downto 0);
           audioRead : in std_logic;
           audioWrite : in std_logic;
           encodeInAddr : in std_logic_vector(18 downto 0);
           encodeOutAddr : in std_logic_vector(18 downto 0);
           encodeInData : out std_logic_vector(31 downto 0);
           encodeOutData : in std_logic_vector(31 downto 0);
           encodeRead : in std_logic;
           encodeWrite : in std_logic;
           inAddress : out std_logic_vector(18 downto 0);
           inData : in std_logic_vector(31 downto 0);
           outData : out std_logic_vector(31 downto 0);
           outAddress : out std_logic_vector(18 downto 0);
           read : out std_logic;
           write : out std_logic);
end component;



signal GND : STD_LOGIC;
signal GND19 : STD_LOGIC_VECTOR(18 downto 0);
signal GND32 : STD_LOGIC_VECTOR(31 downto 0);
signal VDD : STD_LOGIC;

-- invert the active low switches
signal invRecordStart : STD_LOGIC;
signal invPlayStart : STD_LOGIC;
signal invEncode : STD_LOGIC;

-- signals to RAM 
signal read : STD_LOGIC;
signal write : STD_LOGIC;
signal outAddress : STD_LOGIC_VECTOR(18 downto 0);
signal inAddress : STD_LOGIC_VECTOR(18 downto 0);
signal outData : STD_LOGIC_VECTOR(31 downto 0);
signal inData : STD_LOGIC_VECTOR(31 downto 0);

-- signals from audio
signal audioread : STD_LOGIC;
signal audiowrite : STD_LOGIC;
signal playAddress : STD_LOGIC_VECTOR(18 downto 0);
signal recordAddress : STD_LOGIC_VECTOR(18 downto 0);
signal playData : STD_LOGIC_VECTOR(31 downto 0);
signal recordData : STD_LOGIC_VECTOR(31 downto 0);

-- signals from encoder 

signal encoderead : STD_LOGIC;
signal encodewrite : STD_LOGIC;
signal encodeinaddr : STD_LOGIC_VECTOR (18 downto 0);
signal encodeoutaddr : STD_LOGIC_VECTOR(18 downto 0);
signal encodeindata : STD_LOGIC_VECTOR(31 downto 0);
signal encodeoutdata : STD_LOGIC_VECTOR(31 downto 0);

-- signal from ram to encoder

signal canread : STD_LOGIC;
signal canwrite : STD_LOGIC;

begin
	GND <= '0';
	VDD <= '1';
	GND19 <= (others => '0');
	GND32 <= (others => '0');

	-- switches are active low so invert them
	invRecordStart <= not recordStart;
	invPlayStart <= not playStart;
	invEncode <= not encode;

	
	themux : encodemux port map (
	   playStart => invPlayStart,
		recordStart => invRecordStart,
		reset => rstn,
		encode => invEncode,
		playAddress => playAddress,
		recordAddress => recordAddress,
		playData => playData,
		recordData => recordData,
		audioRead => audioRead,
		audioWrite => audioWrite,
		encodeInAddr => encodeInAddr,
		encodeOutAddr => encodeOutAddr,
		encodeInData => encodeInData,
		encodeOutData => encodeOutData,
		encodeRead => encodeRead,
		encodeWrite => encodeWrite,
		inAddress => inAddress,
		inData => inData,
		outData => outData,
		outAddress => outAddress,
		read => read,
		write => write
	);

	audioDevice : audio port map(
        clk => clk,
        rstn => rstn,
        sdout => sdout,
        recordStart => invRecordStart,
        recordStop => GND,
        playStart => invPlayStart,
        playStop => GND,
		direction => direction,
		speed => speed,
		precision => precision,
		volume => volume,
		dataIn => playData,
        mclk => mclk,
        lrck => lrck,
        sclk => sclk,
        sdin => sdin,
        recordAddress => recordAddress,
        playAddress => playAddress,
        dataOut => recordData,
        read => audioread,
        write => audiowrite,
        full => full,
        done => done
    );


	RAM : sraminterface port map(
        CLK => clk,
        Resetn => rstn,
        doRead => read,
        doWrite => write,
        readAddr => inAddress,
        writeAddr => outAddress,
        readData => inData,
        writeData => outData,
        canRead => canread,
        canWrite => canwrite,
        CELeftn => CELeftn,
        CERightn => CERightn,
        OELeftn => OELeftn,
        OERightn => OERightn,
        WELeftn => WELeftn,
        WERightn => WERightn,
        SRAMLeftAddr => SRAMLeftAddr,
        SRAMRightAddr => SRAMRightAddr,
        SRAMLeftData => SRAMLeftData,
        SRAMRightData => SRAMRightData
    );


	 theencoder : encoder port map (
			  read => encoderead,
           write => encodewrite,
			  inaddr => encodeInAddr,
	        outaddr => encodeOutAddr,
           indata => encodeInData,
           outdata => encodeOutData,
			  canread => canread,
			  canwrite => canwrite,
           encode => invEncode,
			  clk => clk,
			  reset => rstn,
			  encodedone => encodedone,
			  doingstuff => doingstuff
	 );

end audiotop_structure;
