library IEEE, SYNOPSYS;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.NUMERIC_STD.ALL;     -- Adam included for TO_SIGNED

entity encoder is
   Port ( 
	        read : out std_logic;
			  write : out std_logic;
	        inaddr : out std_logic_vector(18 downto 0);
	        outaddr : out std_logic_vector(18 downto 0);
           indata : in std_logic_vector(31 downto 0);  	-- wave sound file
           outdata : out std_logic_vector(31 downto 0);	-- ADPCM encrypted
           canread : in std_logic;
			  canwrite : in std_logic;
           encode : in std_logic;
	 		  clk : in std_logic;
			  reset : in std_logic;
			  encodedone : out std_logic;
			  doingstuff : out std_logic);			  		  
end encoder;

---------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------
-------------------------------- BEGINNING OF ENCODER RTL -----------------------------------------------
---------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------

architecture rtl of encoder is

	--------buffer signals-------------
	signal inputbuffer : std_logic_vector(31 downto 0);
	signal outputbuffer: std_logic_vector(31 downto 0);


	-------types-----------------------

	type stateType is (swait, init, waitread, readstate, postram, step1a, step1b, negativediff, step2a, step2b, step2c,
				step2if3withstep3, step4, step5a, step5b, step5cwithstep6, waitwrite, stepwrite, upaddr, fixbuf, flushbuf);

	type indextabletype is array (15 downto 0) of integer;							
	type stepsizetabletype is array (88 downto 0) of integer;						

	------ctrl signals-----------------

	signal state : stateType := swait;	
	signal nextstate : stateType := swait;	

	-------var signals-----------------
	
	signal indextable : indextabletype;			
	signal stepsizetable : stepsizetabletype;	

  	signal sign : bit;				
	signal delta : std_logic_vector (3 downto 0);	
	signal step : std_logic_vector (15 downto 0);	
	signal valpred : integer;			
	signal index : integer;
	signal bufferstep : bit_vector(2 downto 0);			
	signal val : std_logic_vector (15 downto 0);	
	signal diff : integer;					
	signal vpdiff : integer;
	signal localinaddr : std_logic_vector(18 downto 0);
	signal localoutaddr : std_logic_vector(18 downto 0);	
	signal inhalf : bit;

begin
   
	indextable <= (-1, -1, -1, -1, 2, 4, 6, 8,						
	               -1, -1, -1, -1, 2, 4, 6, 8);
	  
	stepsizetable <= (7, 8, 9, 10, 11, 12, 13, 14, 16, 17,				
					19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
					50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
					130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
					337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
					876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066, 
					2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358, 
					5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
					15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767);


   --------------------------------------------------------
	-----------change our state ----------------------------
   --------------------------------------------------------
	process (clk, reset, nextstate)
	begin
		if (reset = '0') then
			state <= swait;
		elsif (clk'EVENT and clk = '1') then
			state <= nextstate;
	   end if;
	end process;

	--------------------------------------------------------
 	------------------next state----------------------------

	controller_ns : process (state)	-- fills in next state
	begin											
		case state is 						
			when swait => doingstuff <= '1';
							  if (encode = '0') then 
							  		nextstate <= swait;
							  else
							  		nextstate <= init; 
							  end if;	
			when init => nextstate <= waitread;
			when waitread => if (canread = '1') then nextstate <= readstate; end if;	
			when readstate => nextstate <= postram;
			when postram => nextstate <= step1a;
			when step1a => nextstate <= step1b;
			when step1b => if (sign = '0') then		
									nextstate <= step2a;		
								else 
									nextstate <= negativediff;	
								end if;			
			when negativediff => nextstate <= step2a;			
			when step2a => nextstate <= step2b;				
			when step2b => nextstate <= step2c;				
			when step2c => nextstate <= step2if3withstep3;		
			when step2if3withstep3 => nextstate <= step4;		
			when step4 => nextstate <= step5a;				
			when step5a => nextstate <= step5b;				
			when step5b => nextstate <= step5cwithstep6;		
			when step5cwithstep6 => 
							if (bufferstep = "111") then 		
								nextstate <= waitwrite;	
							else 
								if (inhalf = '0') then
									nextstate <= waitread;
 							   else
									nextstate <= postram;
								end if;
							end if;			
			when waitwrite => if (canwrite = '1') then nextstate <= stepwrite;	end if;
			when stepwrite => nextstate <= upaddr;			
			when upaddr => 
							if (localinaddr(18) = '0') then 		
								nextstate <= fixbuf;	
							else
 						   	if (inhalf = '0') then
									nextstate <= waitread;
 							   else
									nextstate <= postram;
								end if;
							end if;
			when fixbuf => 
							if(bufferstep = "000") then
								nextstate <= swait;
								encodedone <= '1';
							else
								nextstate <= flushbuf;
							end if;
			when flushbuf => nextstate <= swait;	
								  encodedone <= '1';
		end case;						
	end process;								


	-----------------------------------------------------------
	----------datapath + output function-----------------------

	controller_op : process (state)
	begin
	
		write <= '0';  -- make sure we don't touch memory by default
		read <= '0';   -- ''

		case state is
			when swait => 
							index <= 0;
							valpred <= 0;

			when init =>
				inputbuffer <= (others => '0');
				outputbuffer <= (others => '0');
				step <= conv_std_logic_vector(stepsizetable(index), 16);
				bufferstep <= "000";
				inhalf <= '0';
				localinaddr(17 downto 0) <= (others => '0');
				localinaddr(18) <= '1';
				localoutaddr <= (others => '0');
			
			when waitread => -- nothing

			when readstate => 
						read <= '1';
						inaddr <= localinaddr;
						inputbuffer <= indata;

			when postram =>
						case inhalf is
							when '0' =>
									-- inhalf is high
									-- have just read
									inhalf <= '1';
									val <= inputbuffer(31 downto 16);
							when '1' =>
									-- input half is low
									-- will read next time
									localinaddr <= localinaddr + 1;
									inhalf <= '0';
									val <= inputbuffer(15 downto 0);
						end case;

			when step1a =>	
						diff <= conv_integer(val) - valpred;

			when step1b => 		
						if (diff < 0) then
							sign <= '1';
						else
							sign <= '0';
						end if;
						
			when negativediff => 	
						diff <= 0 - diff ;
							
			when step2a => 		
						delta <= "0000";
						vpdiff <= conv_integer("0000" & step(15 downto 3));

			when step2b =>		
						if (diff >= conv_integer(step)) then
							delta(2) <= '1';
							diff <= diff - conv_integer(step);
							vpdiff <= vpdiff + conv_integer(step);	
						end if;
						step <= ("0" & step(15 downto 1));

			when step2c =>		
						if (diff >= conv_integer(step)) then
							delta(1) <= '1';							
							diff <= diff - conv_integer(step);							
							vpdiff <= vpdiff + conv_integer(step);	
						end if;
						step <= ("0" & step(15 downto 1));
						
			when step2if3withstep3 => 	
						if (diff >= conv_integer(step)) then
							delta(0) <= '1';														
							vpdiff <= vpdiff + conv_integer(step);	
						end if;
						if(sign ='1') then
							valpred <= valpred - vpdiff;				
						else
							valpred <= valpred + vpdiff;				
						end if;								
		
			when step4 =>
						if( valpred > 32767) then
							valpred <= 32767; 
						elsif (valpred < -32768) then
							valpred <= -32768;
						end if;

			when step5a => 		
						if(sign = '1') then
							delta(3) <= '1';
						end if;
						index <= index + indextable(conv_integer(delta));

			when step5b =>
						if (index < 0) then
							index <= 0;
						elsif (index > 88) then
							index <= 88;
						end if;

			when step5cwithstep6 => 
						step <= conv_std_logic_vector(stepsizetable(index), 16);						
						
						case bufferstep is 
							when "000" => 
											bufferstep <= "001";
											outputbuffer(31 downto 28) <= delta;
							when "001" => 
											bufferstep <= "010";
											outputbuffer(27 downto 24) <= delta;
							when "010" => 
											bufferstep <= "011";
											outputbuffer(23 downto 20) <= delta;
							when "011" => 
											bufferstep <= "100";
											outputbuffer(19 downto 16) <= delta;
							when "100" => 
											bufferstep <= "101";
											outputbuffer(15 downto 12) <= delta;
							when "101" => 
											bufferstep <= "110";
											outputbuffer(11 downto 8) <= delta;
							when "110" => 
											bufferstep <= "111";
											outputbuffer(7 downto 4) <= delta;
							when "111" => 
											bufferstep <= "000";
											outputbuffer(3 downto 0) <= delta;
						end case;
						
			when waitwrite => -- nothing

			when stepwrite =>
						write <= '1';
						outaddr <= localoutaddr;
						outdata <= outputbuffer;
						
			when upaddr =>
						localoutaddr <= localoutaddr + 1;

			when fixbuf =>
						case bufferstep is 
							when "000" => -- do nothing...no more to write
							when "001" => 
											outputbuffer(27 downto 0) <= (others => '0');
							when "010" => 
											outputbuffer(23 downto 0) <= (others => '0');
							when "011" => 
											outputbuffer(19 downto 0) <= (others => '0');
							when "100" => 
											outputbuffer(15 downto 0) <= (others => '0');
							when "101" => 
											outputbuffer(11 downto 0) <= (others => '0');
							when "110" => 
											outputbuffer(7 downto 0) <= (others => '0');
							when "111" => 
											outputbuffer(3 downto 0) <= (others => '0');
						end case;

			when flushbuf =>
						write <= '1';
						outaddr <= localoutaddr;
						outdata <= outputbuffer;

		end case;
	end process;
end rtl;
